function [f,mtf,x,esf,esfp,y,xls,lsf,xlp,lsfp] = ESFreg(x,y,deg,iknots)
    %Estimates Edge Spread Function and Line Spread Function by monotonic
    %regression of raw intensity y in DN projected onto normal x in px,
    %x and trend of y increasing, edge somewhat centered (otherwise fix roi).
    %Algorithms: PAVA and B-Splines with increasing coefficients = I-Splines
    %Splines require bsplinebasis.mexw64, optionally specify a degree (e.g.
    %4) and innerknots (e.g. one at every pixel), see defaults.
    %For spline code see Jan de Leeuw https://rpubs.com/deleeuw/268327
    %By Jack Hogan, Sept 2020. Updated Dec 2024, v0.11:
    %https://www.strollswithmydog.com/spline-esf-lsf-mtf
    
    roi = 16;                                   %+/- px, desired esf range
    idx = find(x>-roi & x<roi);
    x = x(idx); y = y(idx);                     %select data in roi
    x = x(:); y = double(y(:));
    
    if nargin < 4
        iknots = [0.25:0.25:3 3.5:0.5:7 8:roi]; %empirical innerknots
        if nargin < 3
            deg = 4;                            %esf spline degree
        end
    end
    iknots = [-fliplr(iknots) 0 iknots];
    
    %% B-Spline Regression
    
    disp('Fitting ESF with B-Splines + monotone coefficients = I-Splines')
    [h,k]= BsplineBases(x,deg,iknots);          %construct basis functions
    n = size(h,2);                              %d is monotonicity constraint
    d = diag( ones(n,1) ) - diag( ones(n-1,1),1 ); d = d(1:end-1,:);
    u = quadprog(h'*h,-h'*y,d,zeros(n-1,1));    %solve for monotone coeffs
%     u = lsqlin(h,y,d,zeros(n-1,1));             %equivalent but slower
    
    esf = h * u;                                %regressed spline curve
    
    %derivative, regular or scattered data
    xls = (-roi*2:1/16:roi*2-1/16)';            %arbitrary sampling interval
    hp = BsplineBases(xls,deg-1,iknots);        %deg-1 for derivative bases
    up = derivativeCoeffs(u,k,deg);             %coefficients derived from u
    
    lsf = hp * up;                              %spline function derivative
    
    %MTF
    f = 1/(xls(end)-xls(1)) * (0:length(xls)-1)'; %c/p
    
    mtf = abs(fft(lsf));
    mtf = mtf / mtf(1);                         %spline MTF
    mtf50 = interp1(mtf(1:round(n/2)),f(1:round(n/2)),0.5);
        
    %% ESF by PAVA Regression (no LSF, requires binning and corrections)
    
    disp('Fitting ESF with lsqisotonic PAVA')
    w = 1 ./ y;                                 %weights = 1/variance
    esfp = lsqisotonic(x,y,w);                  %monotonic PAVA regression
    
    %finite difference, scattered data
    lsfp = diff(esfp) ./ diff(x);               %PAVA ESF finite difference
    xlp = ( x(2:end) + x(1:end-1) ) / 2;        %shift axis for lsf
    
    %% Output
    
    disp(['B-Spline MTF50 = ' num2str(mtf50) ' c/p'])
    
    figure; subplot(1,3,1); 
    plot(x,y,'.','Color',[0,0.97,0.97]); hold on; 
    plot(x,esfp,x,esf,'k'); grid; grid minor
    xlabel('pixel'); ylabel('Intensity'); xlim([-4 4])
    title ('ESF'); legend('Raw data','PAVA','B-Spline','Location','NW')
    
    subplot(1,3,2); plot(xls,lsf,'.-k')
    xlabel('pixel'); ylabel('Intensity'); xlim([-4 4])
    title ('LSF'); legend('B-Spline Derivative')
    grid; grid minor

    subplot(1,3,3); plot(f,mtf); title('MTF by B-Spline Function Derivative')
    xlim([0 1.5]); grid; grid minor; xlabel('c/p'); ylabel('SFR')
    %%
    function up = derivativeCoeffs(u,k,deg)
        %use spline function coefficients u, knots k and spline degree
        %to derive derivative coefficients up
        %https://pages.mtu.edu/~shene/COURSES/cs3621/NOTES/spline/B-spline/bspline-derv.html
        up = zeros(length(u)-1,1);              %coeffs after derivative
        for i = 1:length(u)-1                   %starting at 1
            up(i) = deg / ( k(i+deg+1)-k(i+1) ) * ( u(i+1) - u (i) );
        end
        up(isnan(up)) = 0; up(isinf(up)) = 0;
    end
end

%%
function [h,knots] = BsplineBases(x,degree,innerknots)
    %B-spline bases for given data, Nx1 vectors,no multiplicities
    %Requires bsplinebasis1.mexw64, compile Jan de Leeuw c source below
    %rpubs.com/deleeuw/268327 & deleeuwpdx.net/pubfolders/splines/splines.pdf
    
    order = degree + 1;
    x = x(:); innerknots = innerknots(:);
    k = isnan(x) | isinf(x);
    if any(k)
        x(k) = [];                      %clean up
    end
    
    lowKnot = min([x(1); innerknots]);
    hiKnot = max([x(end); innerknots]);
    
    knots = [ones(order,1)*lowKnot; innerknots; ones(order,1)*hiKnot];
    nx = length(x);
    nk = length(knots);
    nm = nx * (nk - order);
    
    h = bsplinebasis(x,knots,order,nk,nx,nm);
    h = reshape(h,nx,[]);               %B-spline basis functions
    
    %%      c source code for bsplinebasis.mex64
    %       uncomment, save as isplinebasis.c and compile via mex
    % /*
    %  * bsplinebasis.c - generates spline basis given values of x, knots, order,
    %  *                  number of knots, number of values, and n*m
    %  *
    %  * This is a MEX file for MATLAB required by monoSplineFit.m
    %  *
    %  * The calling syntax is:
    %  *
    %  *		h = isplinebasis(x,knots,order,nk,nx,nm);
    %  *
    %  * Source: Computing and Fitting Monotone Splines
    %  * Jan de Leeuw, April 17, 2017
    %  * http://deleeuwpdx.net/pubfolders/splines/splines.html
    %  * Mex wrapper by Jack Hogan, April 15, 2020
    % */
    %
    % #include "mex.h"
    % #include <math.h>
    % #include <stdbool.h>
    % #include <stdlib.h>
    %
    % inline int VINDEX(const int);
    % inline int MINDEX(const int, const int, const int);
    %
    % void bisect(const double *, const double *, const int *, const int *, int *);
    % void bsplines(const double *, const double *, const int *, const int *, int *,
    %               double *);
    % void bsplineBasis(const double *, const double *, const int *, const int *,
    %                   const int *, double *);
    % void isplineBasis(const double *, const double *, const int *, const int *,
    %                   const int *, double *);
    %
    % inline int VINDEX(const int i) { return i - 1; }
    %
    % inline int MINDEX(const int i, const int j, const int n) {
    %     return (i - 1) + (j - 1) * n;
    % }
    %
    % inline int IMIN(const int a, const int b) {
    %     if (a > b) return b;
    %     return a;
    % }
    %
    % inline int IMAX(const int a, const int b) {
    %     if (a < b) return b;
    %     return a;
    % }
    %
    
    % void bisect(const double *x, const double *knots, const int *lowindex,
    %             const int *highindex, int *index) {
    %     int l = *lowindex, u = *highindex, mid = 0;
    %     while ((u - l) > 1) {
    %         mid = (int)floor((u + l) / 2);
    %         if (*x < knots[VINDEX(mid)])
    %             u = mid;
    %         else
    %             l = mid;
    %     }
    %     *index = l;
    %     return;
    % }
    %
    % void bsplines(const double *x, const double *knots, const int *order,
    %               const int *nknots, int *index, double *q) {
    %     int lowindex = 1, highindex = *nknots, m = *order, j, jp1;
    %     double drr, dll, saved, term;
    %     double *dr = (double *)calloc((size_t)m, sizeof(double));
    %     double *dl = (double *)calloc((size_t)m, sizeof(double));
    %     (void)bisect(x, knots, &lowindex, &highindex, index);
    %     int l = *index;
    %     for (j = 1; j <= m; j++) {
    %         q[VINDEX(j)] = 0.0;
    %     }
    %     if (*x == knots[VINDEX(*nknots)]) {
    %         q[VINDEX(m)] = 1.0;
    %         return;
    %     }
    %     q[VINDEX(1)] = 1.0;
    %     j = 1;
    %     if (j >= m) return;
    %     while (j < m) {
    %         dr[VINDEX(j)] = knots[VINDEX(l + j)] - *x;
    %         dl[VINDEX(j)] = *x - knots[VINDEX(l + 1 - j)];
    %         jp1 = j + 1;
    %         saved = 0.0;
    %         for (int r = 1; r <= j; r++) {
    %             drr = dr[VINDEX(r)];
    %             dll = dl[VINDEX(jp1 - r)];
    %             term = q[VINDEX(r)] / (drr + dll);
    %             q[VINDEX(r)] = saved + drr * term;
    %             saved = dll * term;
    %         }
    %         q[VINDEX(jp1)] = saved;
    %         j = jp1;
    %     }
    %     free(dr);
    %     free(dl);
    %     return;
    % }
    %
    % void bsplineBasis(const double *x, const double *knots, const int *order,
    %                   const int *nknots, const int *nvalues, double *result) {
    %     int m = *order, l = 0;
    %     double *q = (double *)calloc((size_t)m + 1, sizeof(double));
    %     for (int i = 1; i <= *nvalues; i++) {
    %         (void)bsplines(x + VINDEX(i), knots, order, nknots, &l, q);
    %         for (int j = 1; j <= m; j++) {
    %             int r = IMIN(l - m + j, *nknots - m);
    %             result[MINDEX(i, r, *nvalues)] = q[VINDEX(j)];
    %         }
    %     }
    %     free(q);
    %     return;
    % }
    %%
    %
    % /* The gateway function */
    %
    % /* The calling syntax from lsBlineFit.m is:
    %  *
    %  * h = bsplinebasis(x,knots,order,nk,nx,nm);
    %  */
    %
    % void mexFunction( int nlhs, mxArray *plhs[],
    %                   int nrhs, const mxArray *prhs[] )
    % {
    %     double *x, *knots, *h;
    %     int order, nk, nx, nm;
    %
    %     x = mxGetPr(prhs[0]);
    %     knots = mxGetPr(prhs[1]);
    %     order = (int)*mxGetPr(prhs[2]);
    %     nk = (int)*mxGetPr(prhs[3]);    /*number of knots*/
    %     nx = (int)*mxGetPr(prhs[4]);    /*number of values x*/
    %     nm = (int)*mxGetPr(prhs[5]);
    %
    %     printf("order = %d, nk = %d, nx = %d, nm = %d\n",
    %             order, nk, nx, nm);
    %
    %     plhs[0] = mxCreateDoubleMatrix(1,nm,mxREAL);
    %     h = mxGetPr(plhs[0]);
    %
    %     /* Call Computational Function */
    %
    %     bsplineBasis(x,knots,&order,&nk,&nx,h);
    % }
end

function yhat = lsqisotonic(x,y,w)
    %LSQISOTONIC Isotonic least squares.
    %   YHAT = LSQISOTONIC(X,Y) returns a vector of values that minimize the
    %   sum of squares (Y - YHAT).^2 under the monotonicity constraint that
    %   X(I) > X(J) => YHAT(I) >= YHAT(J), i.e., the values in YHAT are
    %   monotonically non-decreasing with respect to X (sometimes referred
    %   to as "weak monotonicity").  LSQISOTONIC uses the "pool adjacent
    %   violators" algorithm.
    %
    %   If X(I) == X(J), then YHAT(I) may be <, ==, or > YHAT(J) (sometimes
    %   referred to as the "primary approach").  If ties do occur in X, a plot
    %   of YHAT vs. X may appear to be non-monotonic at those points.  In fact,
    %   the above monotonicity constraint is not violated, and a reordering
    %   within each group of ties, by ascending YHAT, will produce the desired
    %   appearance in the plot.
    %
    %   YHAT = LSQISOTONIC(X,Y,W) performs weighted isotonic regression using
    %   the non-negative weights in W.
    
    %   Copyright 2003-2006 The MathWorks, Inc.
    
    %   References:
    %      [1] Kruskal, J.B. (1964) "Nonmetric multidimensional scaling: a
    %          numerical method", Psychometrika 29:115-129.
    %      [2] Cox, R.F. and Cox, M.A.A. (1994) Multidimensional Scaling,
    %          Chapman&Hall.
    
    n = numel(x);
    if nargin<3
        yclass = superiorfloat(x,y);
    else
        yclass = superiorfloat(x,y,w);
    end
    
    % Sort points ascending in x, break ties with y.
    [xyord,ord] = sortrows([x(:) y(:)]); iord(ord) = 1:n;
    xyord = double(xyord);
    
    % Initialize fitted values to the given values.
    yhat = xyord(:,2);
    
    block = 1:n;
    if (nargin == 3) && ~isempty(w)
        w = double(w(:)); w = w(ord); % reorder w as a column
        
        % Merge zero-weight points with preceding pos-weighted point (or
        % with the following pos-weighted point if at start).
        posWgts = (w > 0);
        if any(~posWgts)
            idx = cumsum(posWgts); idx(idx == 0) = 1;
            w = w(posWgts);
            yhat = yhat(posWgts);
            block = idx(block);
        end
        
    else
        w = ones(size(yhat));
    end
    
    % Written by Maigo on 8/14/2012 to reduce the complexity from O(n^2) to O(n)
    n = length(yhat);
    b = 0; bstart = zeros(1,n); bend = zeros(1,n);
    for i = 1:n
        b = b + 1;
        yhat(b) = yhat(i);
        w(b) = w(i);
        bstart(b) = i; bend(b) = i;
        while b > 1 && yhat(b) < yhat(b-1)
            yhat(b-1) = (yhat(b-1) * w(b-1) + yhat(b) * w(b)) / (w(b-1) + w(b));
            w(b-1) = w(b-1) + w(b);
            bend(b-1) = bend(b);
            b = b - 1;
        end
    end
    idx = zeros(1,n);
    for i = 1:b
        idx(bstart(i) : bend(i)) = i;
    end
    block = idx(block);
    % Maigo end
    
    % Broadcast merged blocks out to original points, and put back in the
    % original order and shape.
    yhat = yhat(block);
    yhat = reshape(yhat(iord), size(y));
    if isequal(yclass,'single')
        yhat = single(yhat);
    end
end
